import { PublicKey } from '@solana/web3.js';
import { SendAndConfirmTransactionResponse } from '../../rpcModule';
import { TokenMetadataAuthorityHolder, TokenMetadataAuthorityTokenDelegate } from '../Authorization';
import { TransactionBuilder, TransactionBuilderOptions } from '../../../utils';
import { Operation, OperationHandler, Signer, SplTokenAmount } from '../../../types';
import { Metaplex } from '../../../Metaplex';
declare const Key: "DeleteNftOperation";
/**
 * Deletes an existing NFT.
 *
 * ```ts
 * await metaplex
 *   .nfts()
 *   .delete({ mintAddress };
 * ```
 *
 * @group Operations
 * @category Constructors
 */
export declare const deleteNftOperation: import("../../../types").OperationConstructor<DeleteNftOperation, "DeleteNftOperation", DeleteNftInput, DeleteNftOutput>;
/**
 * @group Operations
 * @category Types
 */
export type DeleteNftOperation = Operation<typeof Key, DeleteNftInput, DeleteNftOutput>;
/**
 * @group Operations
 * @category Inputs
 */
export type DeleteNftInput = {
    /** The address of the mint account. */
    mintAddress: PublicKey;
    /**
     * An authority allowed to burn the asset.
     *
     * Note that Metadata authorities are
     * not supported for this instruction.
     *
     * If a `Signer` is provided directly,
     * it will be used as an Holder authority.
     *
     * @see {@link TokenMetadataAuthority}
     * @defaultValue `metaplex.identity()`
     */
    authority?: Signer | TokenMetadataAuthorityTokenDelegate | TokenMetadataAuthorityHolder;
    /**
     * Alias of `authority` for backwards compatibility.
     *
     * @deprecated Use `authority` instead.
     * @see {@link DeleteNftInput.authority}
     */
    owner?: Signer;
    /**
     * The mint of the parent edition when the asset is a printed edition.
     *
     * @defaultValue Defaults to not providing a parent edition to the program.
     */
    parentEditionMint?: PublicKey;
    /**
     * The token account of the parent edition when the asset is a printed edition.
     *
     * @defaultValue Defaults to not providing a parent edition to the program.
     */
    parentEditionToken?: PublicKey;
    /**
     * The edition marker of the asset if it is a printed edition.
     *
     * @defaultValue Defaults to not providing the edition marker to the program.
     */
    editionMarker?: PublicKey;
    /**
     * The explicit token account linking the provided mint and owner
     * accounts, if that account is not their associated token account.
     *
     * @defaultValue Defaults to using the associated token account
     * from the `mintAddress` and `owner` parameters.
     */
    ownerTokenAccount?: PublicKey;
    /**
     * The address of the Sized Collection NFT associated with the
     * NFT to delete, if any. This is required as the collection NFT
     * will need to decrement its size.
     *
     * @defaultValue Defaults to assuming the NFT is not associated with a
     * Size Collection NFT.
     */
    collection?: PublicKey;
    /**
     * The amount of tokens to burn.
     *
     * @defaultValue `token(1)`
     */
    amount?: SplTokenAmount;
};
/**
 * @group Operations
 * @category Outputs
 */
export type DeleteNftOutput = {
    /** The blockchain response from sending and confirming the transaction. */
    response: SendAndConfirmTransactionResponse;
};
/**
 * @group Operations
 * @category Handlers
 */
export declare const deleteNftOperationHandler: OperationHandler<DeleteNftOperation>;
/**
 * @group Transaction Builders
 * @category Inputs
 */
export type DeleteNftBuilderParams = Omit<DeleteNftInput, 'confirmOptions'> & {
    /** A key to distinguish the instruction that burns the NFT. */
    instructionKey?: string;
};
/**
 * Deletes an existing NFT.
 *
 * ```ts
 * const transactionBuilder = metaplex
 *   .nfts()
 *   .builders()
 *   .delete({ mintAddress });
 * ```
 *
 * @group Transaction Builders
 * @category Constructors
 */
export declare const deleteNftBuilder: (metaplex: Metaplex, params: DeleteNftBuilderParams, options?: TransactionBuilderOptions) => TransactionBuilder;
export {};
